@@ -145,7 +145,7 @@ class FeatureVisualization:
             dict_inputs["targets"] = targets
 
             # composites are already registered before
-            self.attribution(data_broadcast, conditions, None)
+            self.attribution(data_broadcast, conditions, None, exclude_parallel=False)
 
             if b % checkpoint == checkpoint - 1:
                 self._save_results((last_checkpoint, sample_indices[-1] + 1))
@@ -306,18 +306,16 @@ class FeatureVisualization:
         mode: "relevance" or "activation"
             Relevance or Activation Maximization 
         r_range: Tuple(int, int)
-            Aange of N-top reference samples. For example, (3, 7) corresponds to the Top-3 to -6 samples.
+            Range of N-top reference samples. For example, (3, 7) corresponds to the Top-3 to -6 samples.
             Argument must be a closed set i.e. second element of tuple > first element.
-        heatmap: boolean
-            If True, compute conditional heatmaps on reference samples. Please make sure to supply a composite.
         composite: zennit.composites or None
-            If `heatmap` is True, `composite` is used for the CondAttribution object.
+            If set, compute conditional heatmaps on reference samples. `composite` is used for the CondAttribution object.
         rf: boolean
-            If True, compute the CRP heatmap for the most relevant neuron only to restrict the conditonal heatmap
+            If True, compute the CRP heatmap for the most relevant/most activating neuron only to restrict the conditonal heatmap
             on the receptive field.
         plot_fn: callable function with signature (samples: torch.Tensor, heatmaps: torch.Tensor, rf: boolean) or None
             Draws reference images. The function receives as input the samples used for computing heatmaps before preprocessing 
-            with self.preprocess and the final heatmaps after computation. In addition, the boolean flag 'rf' is passed to it.
+            with self.preprocess_data and the final heatmaps after computation. In addition, the boolean flag 'rf' is passed to it.
             The return value of the function should correspond to the Cache supplied to the FeatureVisualization object (if available).
             If None, the raw tensors are returned.
         batch_size: int
@@ -369,14 +367,12 @@ class FeatureVisualization:
         mode: "relevance" or "activation"
             Relevance or Activation Maximization 
         r_range: Tuple(int, int)
-            Aange of N-top reference samples. For example, (3, 7) corresponds to the Top-3 to -6 samples.
+            Range of N-top reference samples. For example, (3, 7) corresponds to the Top-3 to -6 samples.
             Argument must be a closed set i.e. second element of tuple > first element.
-        heatmap: boolean
-            If True, compute conditional heatmaps on reference samples. Please make sure to supply a composite.
         composite: zennit.composites or None
-            If `heatmap` is True, `composite` is used for the CondAttribution object.
+            If set, compute conditional heatmaps on reference samples. `composite` is used for the CondAttribution object.
         rf: boolean
-            If True, compute the CRP heatmap for the most relevant neuron only to restrict the conditonal heatmap
+            If True, compute the CRP heatmap for the most relevant/most activating neuron only to restrict the conditonal heatmap
             on the receptive field.
         plot_fn: callable function with signature (samples: torch.Tensor, heatmaps: torch.Tensor, rf: boolean)
             Draws reference images. The function receives as input the samples used for computing heatmaps before preprocessing 
@@ -452,11 +448,12 @@ class FeatureVisualization:
             if rf:
                 neuron_ids = neuron_ids[b * batch_size: (b + 1) * batch_size]
                 conditions = [{layer_name: {concept_id: n_index}} for n_index in neuron_ids]
-                attr = self.attribution(data_batch, conditions, composite, mask_map=ChannelConcept.mask_rf, start_layer=layer_name, on_device=self.device)
+                attr = self.attribution(data_batch, conditions, composite, mask_map=ChannelConcept.mask_rf, start_layer=layer_name, on_device=self.device, 
+                    exclude_parallel=False)
             else:
                 conditions = [{layer_name: [concept_id]}] 
                 # initialize relevance with activation before non-linearity (could be changed in a future release)
-                attr = self.attribution(data_batch, conditions, composite, start_layer=layer_name, on_device=self.device)
+                attr = self.attribution(data_batch, conditions, composite, start_layer=layer_name, on_device=self.device, exclude_parallel=False)
 
             heatmaps.append(attr.heatmap)
 
